// 65PR02 Header File


#define BUF_SIZE		32U				// Amount of Bytes in the Buffer
#define BUF_PBIT		5U				// Amount of Bits the Pointers have

#define MEM_SIZE		1U
#define FPC_INC			MEM_SIZE		// Increment Value for the FPC

// This garbage 0 extends the PC's from 16 to 32 bit and then converts them into signed ints before doing the subtraction
#define BUF_BYTES		(int16_t)((int32_t)((uint32_t)CPU->fetch_pc) - (int32_t)((uint32_t)CPU->reg_pc))

#define CHAR_IN			0xF000
#define CHAR_OUT		0xF000

#define COMBINE(H, L)	(L | (((uint16_t)H) << 8))
#define BTOC(b)			(readflag(CPU, b) ? '1' : '0')
#define FETCH(mem)		if (mem != memtype(FPC)) writebuf(CPU)
#define SETNZ(val)		writeflag(CPU, N_FL, (val) & 128U); writeflag(CPU, Z_FL, !(val))

#define aaa(opc)		((opc >> 5) & 7U)
#define bbb(opc)		((opc >> 2) & 7U)
#define cc(opc)			(opc & 3U)

#define PC				(CPU->reg_pc)
#define FPC				(CPU->fetch_pc)
#define SP				(CPU->reg_sp)
#define A				(CPU->reg_a)
#define X				(CPU->reg_x)
#define Y				(CPU->reg_y)
#define SR				(CPU->reg_sr)

#define C_FL			0U	// Carry
#define Z_FL			1U	// Zero
#define I_FL			2U	// Interrupt
#define D_FL			3U	// Decimal
#define B_FL			4U	// Break
#define U_FL			5U	// Unused
#define V_FL			6U	// Overflow
#define N_FL			7U	// Negative
#define SR_INIT			0b00110100
#define SR_BRK			0b00110000




typedef struct CPU_internal{
	uint16_t reg_pc;				// Program Counter
	uint16_t fetch_pc;				// Fetching Program Counter
	uint16_t buf_wrp : BUF_PBIT;	// Buffer Write Pointer
	uint16_t buf_rdp : BUF_PBIT;	// Buffer Read Pointer
	
	uint8_t mem[65536];				// Main Memory
	uint8_t zp[256];				// Zero Page
	uint8_t stack[256];				// Stack Page
	uint8_t buf[BUF_SIZE];			// Instruction Buffer
	
	uint8_t reg_sp;					// Stack Pointer
	uint8_t reg_a;					// Accumulator
	uint8_t reg_x;					// X Index Register
	uint8_t reg_y;					// Y Index Register
	uint8_t reg_sr;					// Status Register
	
	uint8_t buf_full : 1;			// Buffer Status
} cpu_state;


enum {
	MEM_ZP,
	MEM_STACK,
	MEM_MAIN,
	MEM_NONE
};

enum {
	OP_ORA,		// 000 : ORA
	OP_AND,		// 001 : AND
	OP_XOR,		// 010 : XOR
	OP_ADC,		// 011 : ADC
	OP_Inv0,	// 100 : Invalid
	OP_Inv1,	// 101 : Invalid
	OP_CMP,		// 110 : CMP
	OP_SBC		// 111 : SBC
};

enum {
	OP_ASL,		// 000 : ASL
	OP_ROL,		// 001 : ROL
	OP_LSR,		// 010 : LSR
	OP_ROR,		// 011 : ROR
	OP_Inv2,	// 100 : Invalid
	OP_Inv3,	// 101 : Invalid
	OP_DEC,		// 110 : DEC
	OP_INC		// 111 : INC
};

void g1alu(cpu_state* CPU, uint8_t input, uint8_t operation, bool fetch);
uint8_t g2alu(cpu_state* CPU, uint8_t input, uint8_t operation, bool fetch);

bool readflag(cpu_state* CPU, uint8_t flag);
void writeflag(cpu_state* CPU, uint8_t flag, bool value);

uint8_t pull8(cpu_state* CPU);
uint16_t pull16(cpu_state* CPU);
void push8(cpu_state* CPU, uint8_t value);
void push16(cpu_state* CPU, uint16_t value);

uint8_t memtype(uint16_t address);
uint8_t readmem8(cpu_state* CPU, uint16_t address, bool fetch);
uint16_t readmem16(cpu_state* CPU, uint16_t address, bool wrap, bool fetch);
void writemem8(cpu_state* CPU, uint8_t data, uint16_t address, bool fetch);

int16_t readbuf(cpu_state* CPU);
bool writebuf(cpu_state* CPU);
void resetbuf(cpu_state* CPU);
uint16_t buflen(cpu_state* CPU);

uint32_t readROM(uint8_t** ROM, const char* str);
void init(cpu_state* CPU, uint8_t* initROM, uint32_t ROMsize);
uint8_t execute(cpu_state* CPU, bool print);
void printstate(cpu_state* CPU);


// A:   $00
// X:   $00
// Y:   $00
//SP: $0100     NV--DIZC
//PC: $0000 SR: 00--0000
const char* state_str = " A:   $%02X\n X:   $%02X\n Y:   $%02X\nSP: $01%02X     NV--DIZC\nPC: $%04X SR: %c%c--%c%c%c%c\n";

// 9's Complement Array
const uint8_t ncompl[256] = {
	0x99, 0x98, 0x97, 0x96, 0x95, 0x94, 0x93, 0x92,
	0x91, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90,
	0x89, 0x88, 0x87, 0x86, 0x85, 0x84, 0x83, 0x82,
	0x81, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
	0x79, 0x78, 0x77, 0x76, 0x75, 0x74, 0x73, 0x72,
	0x71, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70,
	0x69, 0x68, 0x67, 0x66, 0x65, 0x64, 0x63, 0x62,
	0x61, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60,
	0x59, 0x58, 0x57, 0x56, 0x55, 0x54, 0x53, 0x52,
	0x51, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50,
	0x49, 0x48, 0x47, 0x46, 0x45, 0x44, 0x43, 0x42,
	0x41, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40,
	0x39, 0x38, 0x37, 0x36, 0x35, 0x34, 0x33, 0x32,
	0x31, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
	0x29, 0x28, 0x27, 0x26, 0x25, 0x24, 0x23, 0x22,
	0x21, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x19, 0x18, 0x17, 0x16, 0x15, 0x14, 0x13, 0x12,
	0x11, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
	0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02,
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02,
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02,
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02,
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02,
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02,
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02,
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

// Instruction Execution cycle counts
const uint8_t inst_cycles[256] = {
	3,	// BRK #		WRST*	RMLP*	WRST
	4,	// ORA (zp,x)	ADCL	RDZP*	RDMM	ALOP
	0,	// 				
	0,	// 				
	3,	// TSB zp		RDZP	ALOP	WRZP
	2,	// ORA zp		RDZP	ALOP
	3,	// ASL zp		RDZP	ALOP	WRZP
	3,	// RMB0 zp		RDZP	ALOP	WRZP
	1,	// PHP			WRST
	1,	// ORA #		ALOP
	1,	// ASL A		ALOP
	0,	// 				
	3,	// TSB abs		RDMM	ALOP	WRMM
	2,	// ORA abs		RDMM	ALOP
	3,	// ASL abs		RDMM	ALOP	WRMM
	2,	// BBR0 zp,r	RDZP	BRCH
	1,	// BPL r		BRCH
	4,	// ORA (zp),y	RDZP*	RDMM	ADCL	ALOP
	3,	// ORA (zp)		RDZP*	RDMM	ALOP
	0,	// 				
	3,	// TRB zp		RDZP	ALOP	WRZP
	3,	// ORA zp,x		ADCL	RDZP	ALOP
	4,	// ASL zp,x		ADCL	RDZP	ALOP	WRZP
	3,	// RMB1 zp		RDZP	ALOP	WRZP
	1,	// CLC			ALOP
	3,	// ORA abs,y	ADCL	RDMM	ALOP
	1,	// INC A		ALOP
	0,	// 				
	3,	// TRB abs		RDMM	ALOP	WRMM
	3,	// ORA abs,x	ADCL	RDMM	ALOP
	4,	// ASL abs,x	ADCL	RDMM	ALOP	WRMM
	2,	// BBR1 zp,r	RDZP	BRCH
	1,	// JSR abs		WSLP*
	4,	// AND (zp,x)	ADCL	RDZP*	RDMM	ALOP
	0,	// 				
	0,	// 				
	2,	// BIT zp		RDZP	ALOP
	2,	// AND zp		RDZP	ALOP
	3,	// ROL zp		RDZP	ALOP	WRZP
	3,	// RMB2 zp		RDZP	ALOP	WRZP
	1,	// PLP			RDST
	1,	// AND #		ALOP
	1,	// ROL A		ALOP
	0,	// 				
	2,	// BIT abs		RDMM	ALOP
	2,	// AND abs		RDMM	ALOP
	3,	// ROL abs		RDMM	ALOP	WRMM
	2,	// BBR2 zp,r	RDZP	BRCH
	1,	// BMI r		BRCH
	4,	// AND (zp),y	RDZP*	RDMM	ADCL	ALOP
	3,	// AND (zp)		RDZP*	RDMM	ALOP
	0,	// 				
	3,	// BIT zp,x		ADCL	RDZP	ALOP
	3,	// AND zp,x		ADCL	RDZP	ALOP
	4,	// ROL zp,x		ADCL	RDZP	ALOP	WRZP
	3,	// RMB3 zp		RDZP	ALOP	WRZP
	1,	// SEC			ALOP
	3,	// AND abs,y	ADCL	RDMM	ALOP
	1,	// DEC A		ALOP
	0,	// 				
	3,	// BIT abs,x	ADCL	RDMM	ALOP
	3,	// AND abs,x	ADCL	RDMM	ALOP
	4,	// ROL abs,x	ADCL	RDMM	ALOP	WRMM
	2,	// BBR3 zp,r	RDZP	BRCH
	1,	// RTI			RDST	RSLP
	4,	// XOR (zp,x)	ADCL	RDZP*	RDMM	ALOP
	0,	// 				
	0,	// 				
	0,	// 				
	2,	// XOR zp		RDZP	ALOP
	3,	// LSR zp		RDZP	ALOP	WRZP
	3,	// RMB4 zp		RDZP	ALOP	WRZP
	1,	// PHA			WRST
	1,	// XOR #		ALOP
	1,	// LSR A		ALOP
	0,	// 				
	1,	// JMP abs		LDPC
	2,	// XOR abs		RDMM	ALOP
	3,	// LSR abs		RDMM	ALOP	WRMM
	2,	// BBR4 zp,r	RDZP	BRCH
	1,	// BVC r		BRCH
	4,	// XOR (zp),y	RDZP*	RDMM	ADCL	ALOP
	3,	// XOR (zp)		RDZP*	RDMM	ALOP
	0,	// 				
	0,	// 				
	3,	// XOR zp,x		ADCL	RDZP	ALOP
	4,	// LSR zp,x		ADCL	RDZP	ALOP	WRZP
	3,	// RMB5 zp		RDZP	ALOP	WRZP
	1,	// CLI			ALOP
	3,	// XOR abs,y	ADCL	RDMM	ALOP
	1,	// PHY			WRST
	0,	// 				
	0,	// 				
	3,	// XOR abs,x	ADCL	RDMM	ALOP
	4,	// LSR abs,x	ADCL	RDMM	ALOP	WRMM
	2,	// BBR5 zp,r	RDZP	BRCH
	1,	// RTS			RSLP*
	4,	// ADC (zp,x)	ADCL	RDZP*	RDMM	ALOP
	0,	// 				
	0,	// 				
	1,	// STZ zp		WRZP
	2,	// ADC zp		RDZP	ALOP
	3,	// ROR zp		RDZP	ALOP	WRZP
	3,	// RMB6 zp		RDZP	ALOP	WRZP
	1,	// PLA			RDST
	1,	// ADC #		ALOP
	1,	// ROR A		ALOP
	0,	// 				
	1,	// JMP (abs)	RMLP*
	2,	// ADC abs		RDMM	ALOP
	3,	// ROR abs		RDMM	ALOP	WRMM
	2,	// BBR6 zp,r	RDZP	BRCH
	1,	// BVS r		BRCH
	4,	// ADC (zp),y	RDZP*	RDMM	ADCL	ALOP
	3,	// ADC (zp)		RDZP*	RDMM	ALOP
	0,	// 				
	2,	// STZ zp,x		ADCL	WRZP
	3,	// ADC zp,x		ADCL	RDZP	ALOP
	4,	// ROR zp,x		ADCL	RDZP	ALOP	WRZP
	3,	// RMB7 zp		RDZP	ALOP	WRZP
	1,	// SEI			ALOP
	3,	// ADC abs,y	ADCL	RDMM	ALOP
	1,	// PLY			RDST
	0,	// 				
	2,	// JMP (abs,x)	ADCL	RMLP*
	3,	// ADC abs,x	ADCL	RDMM	ALOP
	4,	// ROR abs,x	ADCL	RDMM	ALOP	WRMM
	2,	// BBR7 zp,r	RDZP	BRCH
	1,	// BRA r		LDPC
	3,	// STA (zp,x)	ADCL	RDZP*	WRMM
	0,	// 				
	0,	// 				
	1,	// STY zp		WRZP
	1,	// STA zp		WRZP
	1,	// STX zp		WRZP
	3,	// SMB0 zp		RDZP	ALOP	WRZP
	1,	// DEY			ALOP
	1,	// BIT #		ALOP
	1,	// TXA			TRFR
	0,	// 				
	1,	// STY abs		WRMM
	1,	// STA abs		WRMM
	1,	// STX abs		WRMM
	2,	// BBS0 zp,r	RDZP	BRCH
	1,	// BCC r		BRCH
	3,	// STA (zp),y	RDZP*	ADCL	WRMM
	2,	// STA (zp)		RDZP*	WRMM
	0,	// 				
	2,	// STY zp,x		ADCL	WRZP
	2,	// STA zp,x		ADCL	WRZP
	2,	// STX zp,y		ADCL	WRZP
	3,	// SMB1 zp		RDZP	ALOP	WRZP
	1,	// TYA			TRFR
	2,	// STA abs,y	ADCL	WRMM
	1,	// TXS			TRFR
	0,	// 				
	1,	// STZ abs		WRMM
	2,	// STA abs,x	ADCL	WRMM
	2,	// STZ abs,x	ADCL	WRMM
	2,	// BBS1 zp,r	RDZP	BRCH
	1,	// LDY #		TRFR
	4,	// LDA (zp,x)	ADCL	RDZP*	RDMM	ALOP
	1,	// LDX #		TRFR
	0,	// 				
	1,	// LDY zp		RDZP
	1,	// LDA zp		RDZP
	1,	// LDX zp		RDZP
	3,	// SMB2 zp		RDZP	ALOP	WRZP
	1,	// TAY			TRFR
	1,	// LDA #		TRFR
	1,	// TAX			TRFR
	0,	// 				
	1,	// LDY abs		RDMM
	1,	// LDA abs		RDMM
	1,	// LDX abs		RDMM
	2,	// BBS2 zp,r	RDZP	BRCH
	1,	// BCS r		BRCH
	4,	// LDA (zp),y	RDZP*	RDMM	ADCL	ALOP
	3,	// LDA (zp)		RDZP*	RDMM	ALOP
	0,	// 				
	2,	// LDY zp,x		ADCL	RDZP
	2,	// LDA zp,x		ADCL	RDZP
	2,	// LDX zp,y		ADCL	RDZP
	3,	// SMB3 zp		RDZP	ALOP	WRZP
	1,	// CLV			ALOP
	2,	// LDA abs,y	ADCL	RDMM
	1,	// TSX			TRFR
	0,	// 				
	2,	// LDY abs,x	ADCL	RDMM
	2,	// LDA abs,x	ADCL	RDMM
	2,	// LDX abs,y	ADCL	RDMM
	2,	// BBS3 zp,r	RDZP	BRCH
	1,	// CPY #		ALOP
	4,	// CMP (zp,x)	ADCL	RDZP*	RDMM	ALOP
	0,	// 				
	0,	// 				
	2,	// CPY zp		RDZP	ALOP
	2,	// CMP zp		RDZP	ALOP
	3,	// DEC zp		RDZP	ALOP	WRZP
	3,	// SMB4 zp		RDZP	ALOP	WRZP
	1,	// INY			ALOP
	1,	// CMP #		ALOP
	1,	// DEX			ALOP
	1,	// WAI			WAIT
	2,	// CPY abs		RDMM	ALOP
	2,	// CMP abs		RDMM	ALOP
	3,	// DEC abs		RDMM	ALOP	WRMM
	2,	// BBS4 zp,r	RDZP	BRCH
	1,	// BNE r		BRCH
	4,	// CMP (zp),y	RDZP*	RDMM	ADCL	ALOP
	3,	// CMP (zp)		RDZP*	RDMM	ALOP
	0,	// 				
	0,	// 				
	3,	// CMP zp,x		ADCL	RDZP	ALOP
	4,	// DEC zp,x		ADCL	RDZP	ALOP	WRZP
	3,	// SMB5 zp		RDZP	ALOP	WRZP
	1,	// CLD			ALOP
	3,	// CMP abs,y	ADCL	RDMM	ALOP
	1,	// PHX			WRST
	1,	// STP			STOP?
	0,	// 				
	3,	// CMP abs,x	ADCL	RDMM	ALOP
	4,	// DEC abs,x	ADCL	RDMM	ALOP	WRMM
	2,	// BBS5 zp,r	RDZP	BRCH
	1,	// CPX #		ALOP
	4,	// SBC (zp,x)	ADCL	RDZP*	RDMM	ALOP
	0,	// 				
	0,	// 				
	2,	// CPX zp		RDZP	ALOP
	2,	// SBC zp		RDZP	ALOP
	3,	// INC zp		RDZP	ALOP	WRZP
	3,	// SMB6 zp		RDZP	ALOP	WRZP
	1,	// INX			ALOP
	1,	// SBC #		ALOP
	1,	// NOP			NONE
	0,	// 				
	2,	// CPX abs		RDMM	ALOP
	2,	// SBC abs		RDMM	ALOP
	3,	// INC abs		RDMM	ALOP	WRMM
	2,	// BBS6 zp,r	RDZP	BRCH
	1,	// BEQ r		BRCH
	4,	// SBC (zp),y	RDZP*	RDMM	ADCL	ALOP
	3,	// SBC (zp)		RDZP*	RDMM	ALOP
	0,	// 				
	0,	// 				
	3,	// SBC zp,x		ADCL	RDZP	ALOP
	4,	// INC zp,x		ADCL	RDZP	ALOP	WRZP
	3,	// SMB7 zp		RDZP	ALOP	WRZP
	1,	// SED			ALOP
	3,	// SBC abs,y	ADCL	RDMM	ALOP
	1,	// PLX			RDST
	0,	// 				
	0,	// 				
	3,	// SBC abs,x	ADCL	RDMM	ALOP
	4,	// INC abs,x	ADCL	RDMM	ALOP	WRMM
	2	// BBS7 zp,r	RDZP	BRCH
};

// Instruction Size (in Bytes)
const uint8_t inst_size[256] = {
	2,	// BRK #		
	2,	// ORA (zp,x)	
	1,	// 				
	1,	// 				
	2,	// TSB zp		
	2,	// ORA zp		
	2,	// ASL zp		
	2,	// RMB0 zp		
	1,	// PHP			
	2,	// ORA #		
	1,	// ASL A		
	1,	// 				
	3,	// TSB abs		
	3,	// ORA abs		
	3,	// ASL abs		
	3,	// BBR0 zp,r	
	2,	// BPL r		
	2,	// ORA (zp),y	
	2,	// ORA (zp)		
	1,	// 				
	2,	// TRB zp		
	2,	// ORA zp,x		
	2,	// ASL zp,x		
	2,	// RMB1 zp		
	1,	// CLC			
	3,	// ORA abs,y	
	1,	// INC A		
	1,	// 				
	3,	// TRB abs		
	3,	// ORA abs,x	
	3,	// ASL abs,x	
	3,	// BBR1 zp,r	
	3,	// JSR abs		
	2,	// AND (zp,x)	
	1,	// 				
	1,	// 				
	2,	// BIT zp		
	2,	// AND zp		
	2,	// ROL zp		
	2,	// RMB2 zp		
	1,	// PLP			
	2,	// AND #		
	1,	// ROL A		
	1,	// 				
	3,	// BIT abs		
	3,	// AND abs		
	3,	// ROL abs		
	3,	// BBR2 zp,r	
	2,	// BMI r		
	2,	// AND (zp),y	
	2,	// AND (zp)		
	1,	// 				
	2,	// BIT zp,x		
	2,	// AND zp,x		
	2,	// ROL zp,x		
	2,	// RMB3 zp		
	1,	// SEC			
	3,	// AND abs,y	
	1,	// DEC A		
	1,	// 				
	3,	// BIT abs,x	
	3,	// AND abs,x	
	3,	// ROL abs,x	
	3,	// BBR3 zp,r	
	1,	// RTI			
	2,	// XOR (zp,x)	
	1,	// 				
	1,	// 				
	1,	// 				
	2,	// XOR zp		
	2,	// LSR zp		
	2,	// RMB4 zp		
	1,	// PHA			
	2,	// XOR #		
	1,	// LSR A		
	1,	// 				
	3,	// JMP abs		
	3,	// XOR abs		
	3,	// LSR abs		
	3,	// BBR4 zp,r	
	2,	// BVC r		
	2,	// XOR (zp),y	
	2,	// XOR (zp)		
	1,	// 				
	1,	// 				
	2,	// XOR zp,x		
	2,	// LSR zp,x		
	2,	// RMB5 zp		
	1,	// CLI			
	3,	// XOR abs,y	
	1,	// PHY			
	1,	// 				
	1,	// 				
	3,	// XOR abs,x	
	3,	// LSR abs,x	
	3,	// BBR5 zp,r	
	1,	// RTS			
	2,	// ADC (zp,x)	
	1,	// 				
	1,	// 				
	2,	// STZ zp		
	2,	// ADC zp		
	2,	// ROR zp		
	2,	// RMB6 zp		
	1,	// PLA			
	2,	// ADC #		
	1,	// ROR A		
	1,	// 				
	3,	// JMP (abs)	
	3,	// ADC abs		
	3,	// ROR abs		
	3,	// BBR6 zp,r	
	2,	// BVS r		
	2,	// ADC (zp),y	
	2,	// ADC (zp)		
	1,	// 				
	2,	// STZ zp,x		
	2,	// ADC zp,x		
	2,	// ROR zp,x		
	2,	// RMB7 zp		
	1,	// SEI			
	3,	// ADC abs,y	
	1,	// PLY			
	1,	// 				
	3,	// JMP (abs,x)	
	3,	// ADC abs,x	
	3,	// ROR abs,x	
	3,	// BBR7 zp,r	
	2,	// BRA r		
	2,	// STA (zp,x)	
	1,	// 				
	1,	// 				
	2,	// STY zp		
	2,	// STA zp		
	2,	// STX zp		
	2,	// SMB0 zp		
	1,	// DEY			
	2,	// BIT #		
	1,	// TXA			
	1,	// 				
	3,	// STY abs		
	3,	// STA abs		
	3,	// STX abs		
	3,	// BBS0 zp,r	
	2,	// BCC r		
	2,	// STA (zp),y	
	2,	// STA (zp)		
	1,	// 				
	2,	// STY zp,x		
	2,	// STA zp,x		
	2,	// STX zp,y		
	2,	// SMB1 zp		
	1,	// TYA			
	3,	// STA abs,y	
	1,	// TXS			
	1,	// 				
	3,	// STZ abs		
	3,	// STA abs,x	
	3,	// STZ abs,x	
	3,	// BBS1 zp,r	
	2,	// LDY #		
	2,	// LDA (zp,x)	
	2,	// LDX #		
	1,	// 				
	2,	// LDY zp		
	2,	// LDA zp		
	2,	// LDX zp		
	2,	// SMB2 zp		
	1,	// TAY			
	2,	// LDA #		
	1,	// TAX			
	1,	// 				
	3,	// LDY abs		
	3,	// LDA abs		
	3,	// LDX abs		
	3,	// BBS2 zp,r	
	2,	// BCS r		
	2,	// LDA (zp),y	
	2,	// LDA (zp)		
	1,	// 				
	2,	// LDY zp,x		
	2,	// LDA zp,x		
	2,	// LDX zp,y		
	2,	// SMB3 zp		
	1,	// CLV			
	3,	// LDA abs,y	
	1,	// TSX			
	1,	// 				
	3,	// LDY abs,x	
	3,	// LDA abs,x	
	3,	// LDX abs,y	
	3,	// BBS3 zp,r	
	2,	// CPY #		
	2,	// CMP (zp,x)	
	1,	// 				
	1,	// 				
	2,	// CPY zp		
	2,	// CMP zp		
	2,	// DEC zp		
	2,	// SMB4 zp		
	1,	// INY			
	2,	// CMP #		
	1,	// DEX			
	1,	// WAI			
	3,	// CPY abs		
	3,	// CMP abs		
	3,	// DEC abs		
	3,	// BBS4 zp,r	
	2,	// BNE r		
	2,	// CMP (zp),y	
	2,	// CMP (zp)		
	1,	// 				
	1,	// 				
	2,	// CMP zp,x		
	2,	// DEC zp,x		
	2,	// SMB5 zp		
	1,	// CLD			
	3,	// CMP abs,y	
	1,	// PHX			
	1,	// STP			
	1,	// 				
	3,	// CMP abs,x	
	3,	// DEC abs,x	
	3,	// BBS5 zp,r	
	2,	// CPX #		
	2,	// SBC (zp,x)	
	1,	// 				
	1,	// 				
	2,	// CPX zp		
	2,	// SBC zp		
	2,	// INC zp		
	2,	// SMB6 zp		
	1,	// INX			
	2,	// SBC #		
	1,	// NOP			
	1,	// 				
	3,	// CPX abs		
	3,	// SBC abs		
	3,	// INC abs		
	3,	// BBS6 zp,r	
	2,	// BEQ r		
	2,	// SBC (zp),y	
	2,	// SBC (zp)		
	1,	// 				
	1,	// 				
	2,	// SBC zp,x		
	2,	// INC zp,x		
	2,	// SMB7 zp		
	1,	// SED			
	3,	// SBC abs,y	
	1,	// PLX			
	1,	// 				
	1,	// 				
	3,	// SBC abs,x	
	3,	// INC abs,x	
	3	// BBS7 zp,r	
};

#define BRK				0x00
#define ORA_ZPXI		0x01
#define TSB_ZP			0x04
#define ORA_ZP			0x05
#define ASL_ZP			0x06
#define RMB0_ZP			0x07
#define PHP				0x08
#define ORA_I			0x09
#define ASL				0x0A
#define TSB_A			0x0C
#define ORA_A			0x0D
#define ASL_A			0x0E
#define BBR0_ZPR		0x0F
#define BPL_R			0x10
#define ORA_ZPIY		0x11
#define ORA_ZPI			0x12
#define TRB_ZP			0x14
#define ORA_ZPX			0x15
#define ASL_ZPX			0x16
#define RMB1_ZP			0x17
#define CLC				0x18
#define ORA_AY			0x19
#define INC				0x1A
#define TRB_A			0x1C
#define ORA_AX			0x1D
#define ASL_AX			0x1E
#define BBR1_ZPR		0x1F
#define JSR_A			0x20
#define AND_ZPXI		0x21
#define BIT_ZP			0x24
#define AND_ZP			0x25
#define ROL_ZP			0x26
#define RMB2_ZP			0x27
#define PLP				0x28
#define AND_I			0x29
#define ROL				0x2A
#define BIT_A			0x2C
#define AND_A			0x2D
#define ROL_A			0x2E
#define BBR2_ZPR		0x2F
#define BMI_R			0x30
#define AND_ZPIY		0x31
#define AND_ZPI			0x32
#define BIT_ZPX			0x34
#define AND_ZPX			0x35
#define ROL_ZPX			0x36
#define RMB3_ZP			0x37
#define SEC				0x38
#define AND_AY			0x39
#define DEC				0x3A
#define BIT_AX			0x3C
#define AND_AX			0x3D
#define ROL_AX			0x3E
#define BBR3_ZPR		0x3F
#define RTI				0x40
#define XOR_ZPXI		0x41
#define XOR_ZP			0x45
#define LSR_ZP			0x46
#define RMB4_ZP			0x47
#define PHA				0x48
#define XOR_I			0x49
#define LSR				0x4A
#define JMP_A			0x4C
#define XOR_A			0x4D
#define LSR_A			0x4E
#define BBR4_ZPR		0x4F
#define BVC_R			0x50
#define XOR_ZPIY		0x51
#define XOR_ZPI			0x52
#define XOR_ZPX			0x55
#define LSR_ZPX			0x56
#define RMB5_ZP			0x57
#define CLI				0x58
#define XOR_AY			0x59
#define PHY				0x5A
#define XOR_AX			0x5D
#define LSR_AX			0x5E
#define BBR5_ZPR		0x5F
#define RTS				0x60
#define ADC_ZPXI		0x61
#define STZ_ZP			0x64
#define ADC_ZP			0x65
#define ROR_ZP			0x66
#define RMB6_ZP			0x67
#define PLA				0x68
#define ADC_I			0x69
#define ROR				0x6A
#define JMP_AI			0x6C
#define ADC_A			0x6D
#define ROR_A			0x6E
#define BBR6_ZPR		0x6F
#define BVS_R			0x70
#define ADC_ZPIY		0x71
#define ADC_ZPI			0x72
#define STZ_ZPX			0x74
#define ADC_ZPX			0x75
#define ROR_ZPX			0x76
#define RMB7_ZP			0x77
#define SEI				0x78
#define ADC_AY			0x79
#define PLY				0x7A
#define JMP_AXI			0x7C
#define ADC_AX			0x7D
#define ROR_AX			0x7E
#define BBR7_ZPR		0x7F
#define BRA_R			0x80
#define STA_ZPXI		0x81
#define STY_ZP			0x84
#define STA_ZP			0x85
#define STX_ZP			0x86
#define SMB0_ZP			0x87
#define DEY				0x88
#define BIT_I			0x89
#define TXA				0x8A
#define STY_A			0x8C
#define STA_A			0x8D
#define STX_A			0x8E
#define BBS0_ZPR		0x8F
#define BCC_R			0x90
#define STA_ZPIY		0x91
#define STA_ZPI			0x92
#define STY_ZPX			0x94
#define STA_ZPX			0x95
#define STX_ZPY			0x96
#define SMB1_ZP			0x97
#define TYA				0x98
#define STA_AY			0x99
#define TXS				0x9A
#define STZ_A			0x9C
#define STA_AX			0x9D
#define STZ_AX			0x9E
#define BBS1_ZPR		0x9F
#define LDY_I			0xA0
#define LDA_ZPXI		0xA1
#define LDX_I			0xA2
#define LDY_ZP			0xA4
#define LDA_ZP			0xA5
#define LDX_ZP			0xA6
#define SMB2_ZP			0xA7
#define TAY				0xA8
#define LDA_I			0xA9
#define TAX				0xAA
#define LDY_A			0xAC
#define LDA_A			0xAD
#define LDX_A			0xAE
#define BBS2_ZPR		0xAF
#define BCS_R			0xB0
#define LDA_ZPIY		0xB1
#define LDA_ZPI			0xB2
#define LDY_ZPX			0xB4
#define LDA_ZPX			0xB5
#define LDX_ZPY			0xB6
#define SMB3_ZP			0xB7
#define CLV				0xB8
#define LDA_AY			0xB9
#define TSX				0xBA
#define LDY_AX			0xBC
#define LDA_AX			0xBD
#define LDX_AY			0xBE
#define BBS3_ZPR		0xBF
#define CPY_I			0xC0
#define CMP_ZPXI		0xC1
#define CPY_ZP			0xC4
#define CMP_ZP			0xC5
#define DEC_ZP			0xC6
#define SMB4_ZP			0xC7
#define INY				0xC8
#define CMP_I			0xC9
#define DEX				0xCA
#define WAI				0xCB
#define CPY_A			0xCC
#define CMP_A			0xCD
#define DEC_A			0xCE
#define BBS4_ZPR		0xCF
#define BNE_R			0xD0
#define CMP_ZPIY		0xD1
#define CMP_ZPI			0xD2
#define CMP_ZPX			0xD5
#define DEC_ZPX			0xD6
#define SMB5_ZP			0xD7
#define CLD				0xD8
#define CMP_AY			0xD9
#define PHX				0xDA
#define STP				0xDB
#define CMP_AX			0xDD
#define DEC_AX			0xDE
#define BBS5_ZPR		0xDF
#define CPX_I			0xE0
#define SBC_ZPXI		0xE1
#define CPX_ZP			0xE4
#define SBC_ZP			0xE5
#define INC_ZP			0xE6
#define SMB6_ZP			0xE7
#define INX				0xE8
#define SBC_I			0xE9
#define NOP				0xEA
#define CPX_A			0xEC
#define SBC_A			0xED
#define INC_A			0xEE
#define BBS6_ZPR		0xEF
#define BEQ_R			0xF0
#define SBC_ZPIY		0xF1
#define SBC_ZPI			0xF2
#define SBC_ZPX			0xF5
#define INC_ZPX			0xF6
#define SMB7_ZP			0xF7
#define SED				0xF8
#define SBC_AY			0xF9
#define PLX				0xFA
#define SBC_AX			0xFD
#define INC_AX			0xFE
#define BBS7_ZPR		0xFF


const char *inst_names[256] = {
	"BRK #",
	"ORA (zp,x)",
	"INVALID",
	"INVALID",
	"TSB zp",
	"ORA zp",
	"ASL zp",
	"RMB0 zp",
	"PHP",
	"ORA #",
	"ASL A",
	"INVALID",
	"TSB abs",
	"ORA abs",
	"ASL abs",
	"BBR0 zp,r",
	"BPL r",
	"ORA (zp),y",
	"ORA (zp)",
	"INVALID",
	"TRB zp",
	"ORA zp,x",
	"ASL zp,x",
	"RMB1 zp",
	"CLC",
	"ORA abs,y",
	"INC A",
	"INVALID",
	"TRB abs",
	"ORA abs,x",
	"ASL abs,x",
	"BBR1 zp,r",
	"JSR abs",
	"AND (zp,x)",
	"INVALID",
	"INVALID",
	"BIT zp",
	"AND zp",
	"ROL zp",
	"RMB2 zp",
	"PLP",
	"AND #",
	"ROL A",
	"INVALID",
	"BIT abs",
	"AND abs",
	"ROL abs",
	"BBR2 zp,r",
	"BMI r",
	"AND (zp),y",
	"AND (zp)",
	"INVALID",
	"BIT zp,x",
	"AND zp,x",
	"ROL zp,x",
	"RMB3 zp",
	"SEC",
	"AND abs,y",
	"DEC A",
	"INVALID",
	"BIT abs,x",
	"AND abs,x",
	"ROL abs,x",
	"BBR3 zp,r",
	"RTI",
	"XOR (zp,x)",
	"INVALID",
	"INVALID",
	"INVALID",
	"XOR zp",
	"LSR zp",
	"RMB4 zp",
	"PHA",
	"XOR #",
	"LSR A",
	"INVALID",
	"JMP abs",
	"XOR abs",
	"LSR abs",
	"BBR4 zp,r",
	"BVC r",
	"XOR (zp),y",
	"XOR (zp)",
	"INVALID",
	"INVALID",
	"XOR zp,x",
	"LSR zp,x",
	"RMB5 zp",
	"CLI",
	"XOR abs,y",
	"PHY",
	"INVALID",
	"INVALID",
	"XOR abs,x",
	"LSR abs,x",
	"BBR5 zp,r",
	"RTS",
	"ADC (zp,x)",
	"INVALID",
	"INVALID",
	"STZ zp",
	"ADC zp",
	"ROR zp",
	"RMB6 zp",
	"PLA",
	"ADC #",
	"ROR A",
	"INVALID",
	"JMP (a)",
	"ADC abs",
	"ROR abs",
	"BBR6 zp,r",
	"BVS r",
	"ADC (zp),y",
	"ADC (zp)",
	"INVALID",
	"STZ zp,x",
	"ADC zp,x",
	"ROR zp,x",
	"RMB7 zp",
	"SEI",
	"ADC abs,y",
	"PLY",
	"INVALID",
	"JMP (a,X)",
	"ADC abs,x",
	"ROR abs,x",
	"BBR7 zp,r",
	"BRA r",
	"STA (zp,x)",
	"INVALID",
	"INVALID",
	"STY zp",
	"STA zp",
	"STX zp",
	"SMB0 zp",
	"DEY",
	"BIT #",
	"TXA",
	"INVALID",
	"STY abs",
	"STA abs",
	"STX abs",
	"BBS0 zp,r",
	"BCC r",
	"STA (zp),y",
	"STA (zp)",
	"INVALID",
	"STY zp,x",
	"STA zp,x",
	"STX zp,y",
	"SMB1 zp",
	"TYA",
	"STA abs,y",
	"TXS",
	"INVALID",
	"STZ abs",
	"STA abs,x",
	"STZ abs,x",
	"BBS1 zp,r",
	"LDY #",
	"LDA (zp,x)",
	"LDX #",
	"INVALID",
	"LDY zp",
	"LDA zp",
	"LDX zp",
	"SMB2 zp",
	"TAY",
	"LDA #",
	"TAX",
	"INVALID",
	"LDY abs",
	"LDA abs",
	"LDX abs",
	"BBS2 zp,r",
	"BCS r",
	"LDA (zp),y",
	"LDA (zp)",
	"INVALID",
	"LDY zp,x",
	"LDA zp,x",
	"LDX zp,y",
	"SMB3 zp",
	"CLV",
	"LDA abs,y",
	"TSX",
	"INVALID",
	"LDY abs,x",
	"LDA abs,x",
	"LDX abs,y",
	"BBS3 zp,r",
	"CPY #",
	"CMP (zp,x)",
	"INVALID",
	"INVALID",
	"CPY zp",
	"CMP zp",
	"DEC zp",
	"SMB4 zp",
	"INY",
	"CMP #",
	"DEX",
	"WAI",
	"CPY abs",
	"CMP abs",
	"DEC abs",
	"BBS4 zp,r",
	"BNE r",
	"CMP (zp),y",
	"CMP (zp)",
	"INVALID",
	"INVALID",
	"CMP zp,x",
	"DEC zp,x",
	"SMB5 zp",
	"CLD",
	"CMP abs,y",
	"PHX",
	"STP",
	"INVALID",
	"CMP abs,x",
	"DEC abs,x",
	"BBS5 zp,r",
	"CPX #",
	"SBC (zp,x)",
	"INVALID",
	"INVALID",
	"CPX zp",
	"SBC zp",
	"INC zp",
	"SMB6 zp",
	"INX",
	"SBC #",
	"NOP",
	"INVALID",
	"CPX abs",
	"SBC abs",
	"INC abs",
	"BBS6 zp,r",
	"BEQ r",
	"SBC (zp),y",
	"SBC (zp)",
	"INVALID",
	"INVALID",
	"SBC zp,x",
	"INC zp,x",
	"SMB7 zp",
	"SED",
	"SBC abs,y",
	"PLX",
	"INVALID",
	"INVALID",
	"SBC abs,x",
	"INC abs,x",
	"BBS7 zp,r"
};