#include <stdint.h>
#include <stdbool.h>
#include <conio.h>
#include "FixedMath.h"


#define FIX_TH		0x00000100			// Fixed Point Threshold
bool nearlyEqual(fix16_t A, fix16_t B){
	if (A == B){
		return true;					// Quick return
	} else if (xabs(A - B) <= FIX_TH){	// If that didn't work, check if the difference between A and B is smaller than the Threshold
		return true;
	}
	return false;						// If neither were true, return false
}

#define iXmax		300					// Terminal Width
#define iYmax		100					// Terminal Height

#define CxMin		0xFFFE2000			// -1.875 (fixed point)
#define CxMax		0x0000E000			//  0.875 (fixed point)
#define CyMin		0xFFFEE000			// -1.125 (fixed point)
#define CyMax		0x00012000			//  1.125 (fixed point)

//#define CxMin		0xFFFF6A8C			// -0.58380126953125 (fixed point)
//#define CxMax		0xFFFFC17F			// -0.2441558837890625 (fixed point)
//#define CyMin		0xFFFF56B6			// -0.661285400390625 (fixed point)
//#define CyMax		0xFFFF7B94			// -0.51727294921875 (fixed point)

#define FIX_CACHE

#define Multiplier	4
#define FIX_ONE		0x00010000			// 1.0 (fixed point)
#define IterMax		(255U*Multiplier)	// Maximum Iteration Count
#define ER			2					// Escape Radius
#define ESC2		((ER*ER)*FIX_ONE)	// Escape Radius ^ 2 (fixed point)
#define PREVLIMIT	4					// Previous Point Limit
#define PREVBM		PREVLIMIT - 1		// Previous Point Bit Mask

#define termcolor(R,G,B) cprintf("\x1B[38;2;%u;%u;%um", (uint8_t)R, (uint8_t)G, (uint8_t)B)
#define termreset() cprintf("\x1B[39m")
#define cls() cprintf("\x1B[2J")

int main(){
	
	#ifdef FIX_CACHE
	uint8_t PrevCount;
	uint16_t cacheIndex;
	fix16_t PrevPointX[PREVLIMIT];
	fix16_t PrevPointY[PREVLIMIT];
	#endif
	
	uint8_t color;
	uint16_t iX, iY, PrevIteration, Iteration;
	fix16_t Cx, Cy, Zx, Zy, Zx2, Zy2;
	fix16_t PixelWidth = xdiv((CxMax - CxMin), itox(iXmax));
	fix16_t PixelHeight = xdiv((CyMax - CyMin), itox(iYmax));
	
	
	
	termreset();	// Reset the Terminal Color
	cprintf("Mandelbrot Set Program! Resolution: %u x %u\n", iXmax, iYmax);
	
	for (iY = 0; iY < iYmax; iY++){
		
		Cy = CyMin + xmul(itox(iY), PixelHeight);		// Cy = CyMin + (iY * PixelHeight)
		
		if (xabs(Cy) < (PixelHeight >> 1)) Cy = 0;
		
		for (iX = 0; iX < iXmax; iX++){
			Cx = CxMin + xmul(itox(iX), PixelWidth);	// Cx = CxMin + (iX * PixelWidth)
			
			Zx = 0;
			Zy = 0;
			Zx2 = 0;	// xmul(Zx, Zx)
			Zy2 = 0;	// xmul(Zy, Zy)
			
			#ifdef FIX_CACHE
			for (cacheIndex = 0; cacheIndex < PREVLIMIT; cacheIndex++){
				PrevPointX[cacheIndex] = 0;
				PrevPointY[cacheIndex] = 0;
				cprintf("");
			}
			#endif
			
			for (Iteration = 0; (Iteration < IterMax) && ((Zx2+Zy2) < ESC2); Iteration++){
				Zy = (xmul(Zx, Zy) << 1) + Cy;		// Zy = (Zx * Zy * 2) + Cy
				Zx = Zx2 - Zy2 + Cx;				// Zx = Zx2 - Zy2 + Cx
				Zx2 = xmul(Zx, Zx);				// Zx2 = Zx * Zx
				Zy2 = xmul(Zy, Zy);				// Zy2 = Zy * Zy
				
				#ifdef FIX_CACHE
				if (Iteration > PREVLIMIT){
					for (PrevCount = 0; PrevCount < PREVLIMIT; PrevCount++){
						if (nearlyEqual(Zx, PrevPointX[PrevCount]) && nearlyEqual(Zy, PrevPointY[PrevCount])){
							Iteration = IterMax;
							goto exitLoop;
						}
					}
				}
				PrevPointX[Iteration & PREVBM] = Zx;
				PrevPointY[Iteration & PREVBM] = Zy;
				#endif
			}
			
			#ifdef FIX_CACHE
			exitLoop:
			#endif
			
			
			
			if (Iteration != PrevIteration){
				color = Iteration / Multiplier;
				termcolor(color, color, color);
			}
			cprintf("#");
			
			PrevIteration = Iteration;
			
		}
	}
	termreset();
	return 0;
}


